@@ -241,5 +241,8 @@
 /* SVG */
 #undef HAVE_SVG
 
+/* WebP */
+#undef HAVE_WEBP
+
 /* Gpm_Event has wdx and wdy */
 #undef HAVE_WDX_WDY
