@@ -73,9 +73,7 @@ static void grx_set_text(void)
 		GrSetMode(GR_default_text);
 		in_graphics_mode = 0;
 		_GrCloseVideoDriver();
-#ifdef DOS
-		dos_restore_screen();
-#endif
+		restore_terminal();
 	}
 }
 
@@ -213,9 +211,7 @@ get_another:
 static unsigned char *set_mode_according_to_param(unsigned char *param)
 {
 	unsigned char e[MAX_STR_LEN];
-#ifdef DOS
-	dos_save_screen();
-#endif
+	save_terminal();
 	if (!param || !*param) {
 		return stracpy(cast_uchar "There is no default videomode.\nUse the switch \"-mode WIDTHxHEIGHTxCOLORS\" to select video mode.\nThe COLORS argument may be 16, 256, 32k, 64k, 16M.\n");
 #if 0
