@@ -23,7 +23,7 @@ static void http_get_header(struct connection *c);
 static void test_restart(struct connection *c);
 static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url);
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url);
-static void add_accept(unsigned char **hdr, int *l);
+static void add_accept(unsigned char **hdr, int *l, struct connection *c);
 static void add_accept_language(unsigned char **hdr, int *l, struct http_connection_info *info);
 static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url, struct connection *c);
 static void add_accept_charset(unsigned char **hdr, int *l, struct http_connection_info *info);
@@ -116,6 +116,164 @@ int get_http_code(unsigned char *head, int *code, int *version)
 	return 0;
 }
 
+static const unsigned char *months[12] = {
+	cast_uchar "Jan",
+	cast_uchar "Feb",
+	cast_uchar "Mar",
+	cast_uchar "Apr",
+	cast_uchar "May",
+	cast_uchar "Jun",
+	cast_uchar "Jul",
+	cast_uchar "Aug",
+	cast_uchar "Sep",
+	cast_uchar "Oct",
+	cast_uchar "Nov",
+	cast_uchar "Dec"
+};
+
+#ifdef HAVE_GMTIME
+static const unsigned char *days[7] = {
+	cast_uchar "Sun",
+	cast_uchar "Mon",
+	cast_uchar "Tue",
+	cast_uchar "Wed",
+	cast_uchar "Thu",
+	cast_uchar "Fri",
+	cast_uchar "Sat",
+};
+#endif
+
+time_t parse_http_date(unsigned char *date)	/* this functions is bad !!! */
+{
+#ifdef HAVE_MKTIME
+	time_t t = 0;
+	time_t offset;
+	/* Mon, 03 Jan 2000 21:29:33 GMT */
+	int y;
+	struct tm tm;
+	memset(&tm, 0, sizeof(struct tm));
+
+	date = cast_uchar strchr(cast_const_char date, ' ');
+	if (!date) return (time_t)-1;
+	date++;
+	if (*date >= '0' && *date <= '9') {
+			/* Sun, 06 Nov 1994 08:49:37 GMT */
+			/* Sunday, 06-Nov-94 08:49:37 GMT */
+		y = 0;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+		tm.tm_mday = (date[0] - '0') * 10 + date[1] - '0';
+		date += 2;
+		if (*date != ' ' && *date != '-') return (time_t)-1;
+		date += 1;
+		for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++)
+			if (!casecmp(date, months[tm.tm_mon], 3)) goto f1;
+		return (time_t)-1;
+		f1:
+		date += 3;
+		if (*date != '-' && *date != ' ') return (time_t)-1;
+		date++;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+		if (date[2] == ' ') {
+				/* Sunday, 06-Nov-94 08:49:37 GMT */
+			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+			if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+			tm.tm_year = (date[0] >= '7' ? 1900 : 2000) + (date[0] - '0') * 10 + date[1] - '0' - 1900;
+			date += 2;
+		} else if (date[2] >= '0' && date[2] <= '9' && date[3] >= '0' && date[3] <= '9') {
+				/* Sun, 06 Nov 1994 08:49:37 GMT */
+			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+			if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+			if (date[2] < '0' || date[2] > '9') return (time_t)-1;
+			if (date[3] < '0' || date[3] > '9') return (time_t)-1;
+			tm.tm_year = (date[0] - '0') * 1000 + (date[1] - '0') * 100 + (date[2] - '0') * 10 + date[3] - '0' - 1900;
+			date += 4;
+		} else return (time_t)-1;
+		if (*date != ' ') return (time_t)-1;
+		date++;
+	} else {
+			/* Sun Nov  6 08:49:37 1994 */
+		y = 1;
+		for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++)
+			if (!casecmp(date, months[tm.tm_mon], 3)) goto f2;
+		return (time_t)-1;
+		f2:
+		date += 3;
+		while (*date == ' ') date++;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		tm.tm_mday = date[0] - '0';
+		date++;
+		if (*date != ' ') {
+			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+			tm.tm_mday = tm.tm_mday * 10 + date[0] - '0';
+			date++;
+		}
+		if (*date != ' ') return (time_t)-1;
+		date++;
+	}
+
+	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+	tm.tm_hour = (date[0] - '0') * 10 + date[1] - '0';
+	date += 2;
+	if (*date != ':') return (time_t)-1;
+	date++;
+	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+	tm.tm_min = (date[0] - '0') * 10 + date[1] - '0';
+	date += 2;
+	if (*date != ':') return (time_t)-1;
+	date++;
+	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+	tm.tm_sec = (date[0] - '0') * 10 + date[1] - '0';
+	date += 2;
+	if (y) {
+		if (*date != ' ') return (time_t)-1;
+		date++;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+		if (date[2] < '0' || date[2] > '9') return (time_t)-1;
+		if (date[3] < '0' || date[3] > '9') return (time_t)-1;
+		tm.tm_year = (date[0] - '0') * 1000 + (date[1] - '0') * 100 + (date[2] - '0') * 10 + date[3] - '0' - 1900;
+		date += 4;
+	}
+	if (*date != ' ' && *date) return (time_t)-1;
+
+	/*debug("%02d:%02d:%02d", tm.tm_hour, tm.tm_min, tm.tm_sec);*/
+#if defined(HAVE_TIMEGM) && !defined(HAIKU)
+	t = timegm(&tm);
+	if (t == (time_t)-1)
+		return (time_t)-1;
+	offset = 0;
+#else
+	t = mktime(&tm);
+	if (t == (time_t)-1)
+		return (time_t)-1;
+	memset(&tm, 0, sizeof(struct tm));
+	tm.tm_year = 80;
+	tm.tm_mday = 1;
+	offset = 315532800 - mktime(&tm);
+#endif
+	return (time_t)((uttime)t + (uttime)offset);
+#else
+	return (time_t)-1;
+#endif
+}
+
+unsigned char *print_http_date(time_t date)
+{
+	static unsigned char str[64];
+#ifdef HAVE_GMTIME
+	struct tm *tm = gmtime(&date);
+	snprintf(cast_char str, sizeof str, "%s %s %2d %02d:%02d:%02d %04d", days[tm->tm_wday], months[tm->tm_mon], tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, tm->tm_year + 1900);
+#else
+	snprintf(cast_char str, sizeof str, "%.0f", (double)date);
+#endif
+	return str;
+}
+
 static struct {
 	char *name;
 	int bugs;
@@ -326,20 +484,28 @@ static void http_send_header(struct connection *c)
 	if (proxy) add_proxy_auth_string(&hdr, &l, c->url);
 	if (!info->https_forward) {
 		test_restart(c);
-		add_referer(&hdr, &l, host, c->prev_url);
-		add_accept(&hdr, &l);
-		add_accept_language(&hdr, &l, info);
-		add_accept_encoding(&hdr, &l, host, c);
-		add_accept_charset(&hdr, &l, info);
+		if (!c->doh) {
+			add_referer(&hdr, &l, host, c->prev_url);
+		}
+		add_accept(&hdr, &l, c);
+		if (!c->doh) {
+			add_accept_language(&hdr, &l, info);
+			add_accept_encoding(&hdr, &l, host, c);
+			add_accept_charset(&hdr, &l, info);
+		}
 		add_dnt(&hdr, &l);
 		add_connection(&hdr, &l, http10, proxy, !info->send_close);
-		add_upgrade(&hdr, &l);
-		add_if_modified(&hdr, &l, c);
-		add_range(&hdr, &l, host, c);
-		add_pragma_no_cache(&hdr, &l, c->no_cache);
-		add_auth_string(&hdr, &l, host);
+		if (!c->doh) {
+			add_upgrade(&hdr, &l);
+			add_if_modified(&hdr, &l, c);
+			add_range(&hdr, &l, host, c);
+			add_pragma_no_cache(&hdr, &l, c->no_cache);
+			add_auth_string(&hdr, &l, host);
+		}
 		add_post_header(&hdr, &l, &post);
-		add_cookies(&hdr, &l, host);
+		if (!c->doh) {
+			add_cookies(&hdr, &l, host);
+		}
 		add_extra_options(&hdr, &l);
 	}
 	add_to_str(&hdr, &l, cast_uchar "\r\n");
@@ -385,7 +551,7 @@ static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url)
 {
 	add_to_str(hdr, l, cast_uchar "User-Agent: ");
 	if (SCRUB_HEADERS) {
-		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 10.0; rv:68.0) Gecko/20100101 Firefox/68.0\r\n");
+		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 10.0; rv:91.0) Gecko/20100101 Firefox/91.0\r\n");
 	} else if (!(*http_options.header.fake_useragent)) {
 
 		/*
@@ -490,9 +656,11 @@ static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigne
 	}
 }
 
-static void add_accept(unsigned char **hdr, int *l)
+static void add_accept(unsigned char **hdr, int *l, struct connection *c)
 {
-	if (SCRUB_HEADERS) {
+	if (c->doh) {
+		add_to_str(hdr, l, cast_uchar "Accept: application/dns-message\r\n");
+	} else if (SCRUB_HEADERS) {
 		add_to_str(hdr, l, cast_uchar "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n");
 	} else {
 		add_to_str(hdr, l, cast_uchar "Accept: */*\r\n");
@@ -991,11 +1159,11 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 #endif
-	if (h != 401 && h != 407) {
+	if (h != 401 && h != 407 && !c->doh) {
 		unsigned char *cookie;
 		unsigned char *ch = head;
 		while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
-			set_cookie(NULL, host, cookie);
+			set_cookie(host, cookie);
 			mem_free(cookie);
 		}
 	}
@@ -1054,9 +1222,11 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	e->http_code = h;
 	if (e->head) mem_free(e->head);
 	e->head = head;
+	if (c->doh)
+		e->expire_time = 1;
 	if ((d = parse_http_header(head, cast_uchar "Expires", NULL))) {
 		time_t t = parse_http_date(d);
-		if (t && e->expire_time != 1) e->expire_time = t;
+		if (t != (time_t)-1 && e->expire_time != 1) e->expire_time = t;
 		mem_free(d);
 	}
 	if ((d = parse_http_header(head, cast_uchar "Pragma", NULL))) {
@@ -1073,9 +1243,8 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 			}
 			if (!casecmp(f, cast_uchar "max-age=", 8)) {
 				if (e->expire_time != 1) {
-					errno = 0;
-					EINTRLOOPX(e->expire_time, time(NULL), (time_t)-1);
-					e->expire_time += atoi(cast_const_char(f + 8));
+					e->expire_time = get_absolute_seconds();
+					e->expire_time = (time_t)((uttime)e->expire_time + (uttime)atoi(cast_const_char(f + 8)));
 				}
 			}
 			while (*f && *f != ',') f++;
@@ -1195,12 +1364,15 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 
 	/*
 	 * Truncate entry if:
+	 *	- we are using DNS-over-HTTPS
 	 *	- it is compressed (the mix of an old and new document
 	 *	  would likely produce decompression error).
 	 *	- it was http authentication (the user doesn't need to see the
 	 *	  authentication message).
 	 */
-	if ((d = parse_http_header(e->head, cast_uchar "Content-Encoding", NULL))) {
+	if (c->doh) {
+		truncate_entry(e, c->from, 0);
+	} else if ((d = parse_http_header(e->head, cast_uchar "Content-Encoding", NULL))) {
 		mem_free(d);
 		truncate_entry(e, c->from, 0);
 	} else if (previous_http_code == 401 || previous_http_code == 407) {
