@@ -158,10 +158,8 @@ static void gray_to_rgb(unsigned char *data, int pixels)
 /* Fixes returned data in case they are CMYK or grayscale. */
 static inline void fix_data( struct jpg_decoder *deco, int lines_read)
 {
-#ifdef __ICC
-	volatile	/* ICC bug */
-#endif
-	int a;
+	/* ICC bug */
+	icc_volatile int a;
 
 	switch (global_cinfo->output_components){
 		case 1:
@@ -178,7 +176,7 @@ static inline void fix_data( struct jpg_decoder *deco, int lines_read)
 			->output_width*lines_read);
 		break;
 
-		default: internal("Invalid output_components");
+		default: internal_error("Invalid output_components");
 	}
 }
 
@@ -188,7 +186,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 
 	deco=(struct jpg_decoder *)(cimg->decoder);
 #ifdef DEBUG
-	if (!deco) internal("NULL decoder in jpeg_restart");
+	if (!deco) internal_error("NULL decoder in jpeg_restart");
 #endif /* #ifdef DEBUG */
 	global_cinfo=((struct jpg_decoder *)(cimg->decoder))->cinfo;
 	global_jerr=((struct jpg_decoder *)(cimg->decoder))->jerr;
@@ -250,6 +248,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 			break;
 		global_cinfo->buffered_image=TRUE;
 		deco->state=1;
+		/*-fallthrough*/
 
 		case 1:
 		/* If the scaling is sufficiently brutal we can leave out
@@ -327,6 +326,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 		}
 new_scan:
 		deco->state=2;
+		/*-fallthrough*/
 
 		case 2:
 		/* jpeg_start_output */
@@ -340,26 +340,18 @@ susp0:
 		case 3:
 		/* jpeg_read_scanlines */
 			/* color */
-		while (global_cinfo->output_scanline
-			<global_cinfo->output_height){
+		while (global_cinfo->output_scanline < global_cinfo->output_height) {
 			int a, lines;
 
-			for (a=0;a<16;a++){
-				deco->scanlines[a]=cimg->buffer
-				+(global_cinfo
-				->output_scanline+a)
-				*global_cinfo->output_width*cimg->
-					buffer_bytes_per_pixel;
+			for (a = 0; a < 16; a++) {
+				deco->scanlines[a] = cimg->buffer + ((size_t)global_cinfo->output_scanline + a) * global_cinfo->output_width * cimg->buffer_bytes_per_pixel;
 			}
 
-			if ((lines=
-				jpeg_read_scanlines(
-				global_cinfo,deco->scanlines,1))){
+			if ((lines = jpeg_read_scanlines(global_cinfo, deco->scanlines, 1))) {
 				/* Some lines were written into cimg buffer */
-				cimg->rows_added=1;
-
+				cimg->rows_added = 1;
 				fix_data(deco, lines);
-			}else{
+			} else {
 				/* No lines have been written into cimg
 				 * buffer */
 				/* We are suspended and we want more data */
@@ -368,6 +360,7 @@ susp0:
 			}
 		}
 		deco->state=4;
+		/*-fallthrough*/
 
 		case 4:
 		/* jpeg_finish_output */
@@ -382,6 +375,7 @@ susp0:
 			goto new_scan;
 		}
 		deco->state=5;
+		/*-fallthrough*/
 
 		case 5:
 		/* jpeg_finish_decompress */
@@ -412,14 +406,14 @@ void add_jpeg_version(unsigned char **s, int *l)
 #endif
 #if defined(JPEG_LIB_VERSION_MAJOR) && defined(JPEG_LIB_VERSION_MINOR)
 	add_num_to_str(s, l, JPEG_LIB_VERSION_MAJOR);
-	add_to_str(s, l, cast_uchar ".");
+	add_chr_to_str(s, l, '.');
 	add_num_to_str(s, l, JPEG_LIB_VERSION_MINOR);
 #else
 	add_num_to_str(s, l, JPEG_LIB_VERSION / 10);
-	add_to_str(s, l, cast_uchar ".");
+	add_chr_to_str(s, l, '.');
 	add_num_to_str(s, l, JPEG_LIB_VERSION % 10);
 #endif
-	add_to_str(s, l, cast_uchar ")");
+	add_chr_to_str(s, l, ')');
 }
 
 #endif /* #ifdef HAVE_JPEG */
