@@ -1228,6 +1228,34 @@ static void scroll_list(struct terminal *term, void *bla)
 #endif
 }
 
+static int list_item_mark(struct dialog_data *dlg, struct dialog_item_data *useless)
+{
+	struct list_description *ld = (struct list_description *)dlg->dlg->udata2;
+	static struct redraw_data rd;
+
+	rd.ld = ld;
+	rd.dlg = dlg;
+	rd.n = 0;
+
+	if (ld->current_pos != ld->list) ld->current_pos->type ^= 4;
+	if (next_in_tree(ld, ld->current_pos) == ld->list) /* already at the bottom */
+	{
+		draw_to_window(dlg->win, redraw_list_line, &rd);
+		return 0;
+	}
+	rd.n = -1;
+	ld->current_pos = next_in_tree(ld, ld->current_pos);
+	ld->win_pos++;
+	if (ld->win_pos > ld->n_items - 1)  /* scroll down */
+	{
+		ld->win_pos = ld->n_items - 1;
+		ld->win_offset = next_in_tree(ld, ld->win_offset);
+		draw_to_window(dlg->win, scroll_list, &rd);
+	}
+	draw_to_window(dlg->win, redraw_list_line, &rd);
+	return 0;
+}
+
 
 static void list_find_next(struct redraw_data *rd, int direction)
 {
@@ -1385,22 +1413,7 @@ static int list_event_handler(struct dialog_data *dlg, struct links_event *ev)
 		}
 		if (ev->x == 'i' || ev->x == '*' || ev->x == '8' || ev->x == KBD_INS || ev->x == KBD_SELECT)
 		{
-			if (ld->current_pos!=ld->list)ld->current_pos->type^=4;
-			rd.n=-1;
-			if (next_in_tree(ld,ld->current_pos)==ld->list) /* already at the bottom */
-			{
-				draw_to_window(dlg->win,redraw_list_line,&rd);
-				return EVENT_PROCESSED;
-			}
-			ld->current_pos=next_in_tree(ld,ld->current_pos);
-			ld->win_pos++;
-			if (ld->win_pos>ld->n_items-1)  /* scroll down */
-			{
-				ld->win_pos=ld->n_items-1;
-				ld->win_offset=next_in_tree(ld,ld->win_offset);
-				draw_to_window(dlg->win,scroll_list,&rd);
-			}
-			draw_to_window(dlg->win,redraw_list_line,&rd);
+			list_item_mark(dlg, NULL);
 			return EVENT_PROCESSED;
 		}
 		if (ev->x == KBD_DOWN)
@@ -1733,7 +1746,7 @@ static void create_list_window_fn(struct dialog_data *dlg)
 	int n_items;
 	struct redraw_data rd;
 
-	int a=6;
+	int a=7;
 
 	ld->dlg=dlg;
 	if (ld->button_fn)a++;  /* user button */
@@ -1841,7 +1854,7 @@ int create_list_window(
 		ld->dlg=NULL;
 	}
 
-	a=7;
+	a=8;
 	if (ld->button_fn)a++;
 	if (ld->type==1)a++;
 
@@ -1885,19 +1898,23 @@ int create_list_window(
 	d->items[a+2].fn=list_item_edit;
 
 	d->items[a+3].type=D_BUTTON;
-	d->items[a+3].text=TEXT_(T_MOVE);
-	d->items[a+3].fn=list_item_move;
+	d->items[a+3].text=TEXT_(T_SELECT);
+	d->items[a+3].fn=list_item_mark;
 
 	d->items[a+4].type=D_BUTTON;
-	d->items[a+4].text=TEXT_(T_UNSELECT_ALL);
-	d->items[a+4].fn=list_item_unselect;
+	d->items[a+4].text=TEXT_(T_MOVE);
+	d->items[a+4].fn=list_item_move;
 
 	d->items[a+5].type=D_BUTTON;
-	d->items[a+5].gid=B_ESC;
-	d->items[a+5].fn=cancel_dialog;
-	d->items[a+5].text=TEXT_(T_CLOSE);
+	d->items[a+5].text=TEXT_(T_UNSELECT_ALL);
+	d->items[a+5].fn=list_item_unselect;
 
-	d->items[a+6].type=D_END;
+	d->items[a+6].type=D_BUTTON;
+	d->items[a+6].gid=B_ESC;
+	d->items[a+6].fn=cancel_dialog;
+	d->items[a+6].text=TEXT_(T_CLOSE);
+
+	d->items[a+7].type=D_END;
 	do_dialog(term, d, getml(d, NULL));
 	return 0;
 }
