@@ -78,13 +78,6 @@ void tn3270_func(struct session *ses, unsigned char *url)
 	tn_func(ses, url, &tn3270_prog, TEXT_(T_TN3270), TEXT_(T_BAD_TN3270_URL));
 }
 
-void magnet_func(struct session *ses, unsigned char *url)
-{
-	unsigned char *escaped_url = escape_path(url);
-	prog_func(ses->term, &magnet_prog, escaped_url, TEXT_(T_MAGNET));
-	mem_free(escaped_url);
-}
-
 void mms_func(struct session *ses, unsigned char *url)
 {
 	if (check_shell_url(url)) {
@@ -93,3 +86,17 @@ void mms_func(struct session *ses, unsigned char *url)
 	}
 	prog_func(ses->term, &mms_prog, url, TEXT_(T_MMS));
 }
+
+void magnet_func(struct session *ses, unsigned char *url)
+{
+	unsigned char *escaped_url = escape_path(url);
+	prog_func(ses->term, &magnet_prog, escaped_url, TEXT_(T_MAGNET));
+	mem_free(escaped_url);
+}
+
+void gopher_func(struct session *ses, unsigned char *url)
+{
+	unsigned char *escaped_url = escape_path(url);
+	prog_func(ses->term, &gopher_prog, escaped_url, TEXT_(T_GOPHER));
+	mem_free(escaped_url);
+}
