@@ -236,11 +236,13 @@ static void free_connection_data(struct connection *c)
 		c->info = NULL;
 	}
 	clear_connection_timeout(c);
-	if (--active_connections < 0) {
-		internal_error("active connections underflow");
-		active_connections = 0;
+	if (!c->doh) {
+		if (--active_connections < 0) {
+			internal_error("active connections underflow");
+			active_connections = 0;
+		}
 	}
-	if (c->state != S_WAIT) {
+	if (c->state != S_WAIT && !c->doh) {
 		if ((h = is_host_on_list(c))) {
 			if (!--h->conn) {
 				del_from_list(h);
@@ -464,9 +466,8 @@ static int try_to_suspend_connection(struct connection *c, unsigned char *ho)
 	return -1;
 }
 
-int is_noproxy_url(unsigned char *url)
+int is_noproxy_host(unsigned char *host)
 {
-	unsigned char *host = get_host_name(url);
 	if (!proxies.only_proxies) {
 		unsigned char *np = proxies.no_proxy;
 		int host_l = (int)strlen(cast_const_char host);
@@ -478,7 +479,6 @@ int is_noproxy_url(unsigned char *url)
 				goto no_match;
 			if (casecmp(np, host + (host_l - l), l))
 				goto no_match;
-			mem_free(host);
 			return 1;
 no_match:
 			if (!np[l])
@@ -486,10 +486,17 @@ no_match:
 			np += l + 1;
 		}
 	}
-	mem_free(host);
 	return 0;
 }
 
+int is_noproxy_url(unsigned char *url)
+{
+	unsigned char *host = get_host_name(url);
+	int r = is_noproxy_host(host);
+	mem_free(host);
+	return r;
+}
+
 static void run_connection(struct connection *c)
 {
 	struct h_conn *hc;
@@ -522,17 +529,21 @@ static void run_connection(struct connection *c)
 		del_connection(c);
 		return;
 	}
-	if (!(hc = is_host_on_list(c))) {
-		hc = mem_alloc(sizeof(struct h_conn));
-		if (!(hc->host = get_host_name(c->url))) {
-			mem_free(hc);
-			goto s_bad_url;
+	if (c->doh && func != http_func && func != https_func && func != proxy_func)
+		goto s_bad_url;
+	if (!c->doh) {
+		if (!(hc = is_host_on_list(c))) {
+			hc = mem_alloc(sizeof(struct h_conn));
+			if (!(hc->host = get_host_name(c->url))) {
+				mem_free(hc);
+				goto s_bad_url;
+			}
+			hc->conn = 0;
+			add_to_list(h_conns, hc);
 		}
-		hc->conn = 0;
-		add_to_list(h_conns, hc);
+		hc->conn++;
+		active_connections++;
 	}
-	hc->conn++;
-	active_connections++;
 	c->keepalive = 0;
 	c->running = 1;
 	func(c);
@@ -601,17 +612,19 @@ void abort_connection(struct connection *c)
 
 static int try_connection(struct connection *c)
 {
-	struct h_conn *hc = NULL;
-	if ((hc = is_host_on_list(c))) {
-		if (hc->conn >= max_connections_to_host) {
-			if (try_to_suspend_connection(c, hc->host)) return 0;
+	struct h_conn *hc;
+	if (!c->doh) {
+		if ((hc = is_host_on_list(c))) {
+			if (hc->conn >= max_connections_to_host) {
+				if (try_to_suspend_connection(c, hc->host)) return 0;
+				else return -1;
+			}
+		}
+		if (active_connections >= max_connections) {
+			if (try_to_suspend_connection(c, NULL)) return 0;
 			else return -1;
 		}
 	}
-	if (active_connections >= max_connections) {
-		if (try_to_suspend_connection(c, NULL)) return 0;
-		else return -1;
-	}
 	run_connection(c);
 	return 1;
 }
@@ -627,7 +640,7 @@ static void check_queue_bugs(void)
 	cc = 0;
 	foreach(struct connection, d, ld, queue) {
 		int q = getpri(d);
-		cc += d->running;
+		cc += d->running && !d->doh;
 		if (q < p) {
 			if (!ps) {
 				internal_error("queue is not sorted");
@@ -755,7 +768,7 @@ int disallow_url(unsigned char *url, int allow_flags)
 
 /* prev_url is a pointer to previous url or NULL */
 /* prev_url will NOT be deallocated */
-void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat, int pri, int no_cache, int no_compress, int allow_flags, off_t position)
+void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat, int pri, int no_cache, int no_compress, int doh, int allow_flags, off_t position)
 {
 	struct cache_entry *e = NULL;
 	struct connection *c;
@@ -785,7 +798,7 @@ void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat,
 		}
 	}
 #endif
-	if (is_url_blocked(url)) {
+	if (!doh && is_url_blocked(url)) {
 		if (stat) {
 			stat->state = S_BLOCKED_URL;
 			if (stat->end) stat->end(stat, stat->data);
@@ -806,7 +819,7 @@ void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat,
 			goto skip_cache;
 		}
 		if (!aggressive_cache && no_cache > NC_ALWAYS_CACHE) {
-			if (e->expire_time && e->expire_time < time(NULL)) {
+			if (e->expire_time && e->expire_time < get_absolute_seconds()) {
 				if (no_cache < NC_IF_MOD) no_cache = NC_IF_MOD;
 				e->refcount--;
 				goto skip_cache;
@@ -841,6 +854,10 @@ void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat,
 	u = get_proxy(url);
 	foreach(struct connection, c, lc, queue) if (!c->detached && !strcmp(cast_const_char c->url, cast_const_char u)) {
 		if (c->from < position) continue;
+		if (c->doh != doh) {
+			must_detach = 1;
+			break;
+		}
 		if (no_compress && !c->no_compress) {
 			unsigned char *enc;
 			if ((c->state >= S_WAIT && c->state < S_TRANS) || !c->cache) {
@@ -910,6 +927,7 @@ void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat,
 #else
 	c->no_compress = 1;
 #endif
+	c->doh = doh;
 	c->prg.timer = NULL;
 	c->timer = NULL;
 	if (position || must_detach) {
