@@ -9,7 +9,10 @@
 #include <stsdef.h>
 #include <descrip.h>
 #include <dvidef.h>
+#include <ttdef.h>
+#include <tt2def.h>
 #include <lib$routines.h>
+#include <libdef.h>
 #include <libclidef.h>
 
 #if defined(HAVE_OPENSSL) && !defined(OPENSSL_NO_SHA) && !defined(OPENSSL_NO_SHA1)
@@ -26,9 +29,15 @@ struct _iosb {
 struct _generic_64 {
 	unsigned gen64$l_longword[2];
 };
+struct _va_range {
+	void *va_range$ps_start_va;
+	void *va_range$ps_end_va;
+};
 #else
 #include <iosbdef.h>
 #include <gen64def.h>
+#include <vadef.h>
+#include <va_rangedef.h>
 #endif
 
 #include <pthread.h>
@@ -37,6 +46,75 @@ struct _generic_64 {
 #include <sys/ioctl.h>
 #endif
 
+struct tt_mode {
+	unsigned char class;
+	unsigned char mode;
+	unsigned short width;
+	unsigned basic;
+	unsigned extended;
+};
+
+#ifndef OPENVMS_64BIT
+
+#define va_range _va_range
+#define links_expreg	sys$expreg
+#define links_cretva	sys$cretva
+#define links_deltva	sys$deltva
+
+#else
+
+struct va_range {
+	void *va_range$ps_start_va;
+	void *va_range$ps_end_va;
+};
+
+static int links_expreg(size_t pagelets, struct va_range *retadr, unsigned ignore_mode, char ignore_reg)
+{
+	__int64 region_int = VA$C_P2;
+	struct _generic_64 region;
+	unsigned __int64 len;
+	int vms_ret;
+
+	memcpy(&region, &region_int, 8);
+
+	vms_ret = sys$expreg_64(&region, pagelets << 9, 0, 0, &retadr->va_range$ps_start_va, &len);
+	retadr->va_range$ps_end_va = (unsigned char *)retadr->va_range$ps_start_va + len - 1;
+
+	return vms_ret;
+}
+
+static int links_cretva(struct va_range *creadr, struct va_range *retadr, unsigned int ignore_mode)
+{
+	__int64 region_int = VA$C_P2;
+	struct _generic_64 region;
+	unsigned __int64 len;
+	int vms_ret;
+
+	memcpy(&region, &region_int, 8);
+
+	vms_ret = sys$cretva_64(&region, creadr->va_range$ps_start_va, (unsigned char *)creadr->va_range$ps_end_va - (unsigned char *)creadr->va_range$ps_start_va + 1, 0, 0, &retadr->va_range$ps_start_va, &len);
+	retadr->va_range$ps_end_va = (unsigned char *)retadr->va_range$ps_start_va + len - 1;
+
+	return vms_ret;
+}
+
+static int links_deltva(struct va_range *deladr, struct va_range *retadr, unsigned int ignore_mode)
+{
+	__int64 region_int = VA$C_P2;
+	struct _generic_64 region;
+	unsigned __int64 len;
+	int vms_ret;
+
+	memcpy(&region, &region_int, 8);
+
+	vms_ret = sys$deltva_64(&region, deladr->va_range$ps_start_va, (unsigned char *)deladr->va_range$ps_end_va - (unsigned char *)deladr->va_range$ps_start_va + 1, 0, &retadr->va_range$ps_start_va, &len);
+	retadr->va_range$ps_end_va = (unsigned char *)retadr->va_range$ps_start_va + len - 1;
+
+	return vms_ret;
+}
+
+#endif
+
 /* reimplement pipes because sockets + select have too big latency */
 #define VMS_VIRTUAL_PIPE
 
@@ -53,6 +131,7 @@ struct _generic_64 {
 
 #undef read
 #undef write
+#undef pipe
 #undef close
 #undef select
 
@@ -86,7 +165,7 @@ void portable_sleep(unsigned msec)
 
 static $DESCRIPTOR (output_channel_desc, "SYS$OUTPUT:");
 
-int get_terminal_size(int fd, int *x, int *y)
+void get_terminal_size(int *x, int *y)
 {
 	int ret;
 	int x_code = DVI$_DEVBUFSIZ;
@@ -98,7 +177,6 @@ int get_terminal_size(int fd, int *x, int *y)
 	if ($VMS_STATUS_SUCCESS(ret) && result) *x = result;
 	ret = lib$getdvi(&y_code, 0, &output_channel_desc, &result, 0, 0);
 	if ($VMS_STATUS_SUCCESS(ret) && result) *y = result;
-	return 0;
 }
 
 static $DESCRIPTOR (display_desc, "DECW$DISPLAY");
@@ -117,71 +195,18 @@ void get_path_to_exe(void)
 	path_to_exe = cast_uchar "links";
 }
 
-static volatile int io_raw = -1;
-
+static int io_raw = -1;
 static int input_handle = -1;
 
-static unsigned old_ctrl_mask = 0;
-
 static $DESCRIPTOR (input_channel_desc, "SYS$INPUT:");
 static unsigned short input_channel;
 
 static pthread_mutex_t io_mutex;
 static pthread_mutex_t cancel_mutex;
 
-static volatile int must_redraw = 0;
-static volatile int ctrl_c = 0;
-static volatile int ctrl_y = 0;
-
 static volatile struct timeval gettimeofday_pool[GETTIMEOFDAY_POOL];
 static volatile unsigned gettimeofday_clock = 0;
 
-static void vms_key_handler(int mode, void (*ast)(void))
-{
-	struct _iosb iosb;
-	int ret;
-	ret = sys$qiow(0, input_channel, IO$_SETMODE | mode, &iosb, 0, 0, ast, 0, 0, 0, 0, 0);
-	if (!$VMS_STATUS_SUCCESS(ret))
-		vms_fatal_exit("sys$qiow failed: %d", ret);
-	if (!$VMS_STATUS_SUCCESS(iosb.iosb$w_status))
-		vms_fatal_exit("IO$M_CTRLCAST failed: %d", iosb.iosb$w_status);
-}
-
-static void ctrl_c_ast(void)
-{
-	if (io_raw <= 0) return;
-	must_redraw = 1;
-	ctrl_c = 1;
-	sys$cancel(input_channel);
-	vms_key_handler(IO$M_CTRLCAST, ctrl_c_ast);
-}
-
-static void ctrl_y_ast(void)
-{
-	if (io_raw <= 0) return;
-	must_redraw = 1;
-	ctrl_y = 1;
-	sys$cancel(input_channel);
-	vms_key_handler(IO$M_CTRLYAST, ctrl_y_ast);
-	vms_key_handler(IO$M_CTRLCAST, ctrl_c_ast);
-}
-
-static void handle_ast_keys(int h)
-{
-	if (ctrl_c) {
-		ctrl_c = 0;
-		hard_write(h, cast_uchar "\003", 1);
-	}
-	if (ctrl_y) {
-		ctrl_y = 0;
-		hard_write(h, cast_uchar "\031", 1);
-	}
-	if (must_redraw) {
-		must_redraw = 0;
-		hard_write(h, cast_uchar "\014", 1);
-	}
-}
-
 #define INPUT_BUFFER_SIZE	16
 
 static void vms_input_thread(void *n, int h)
@@ -210,7 +235,6 @@ static void vms_input_thread(void *n, int h)
 		ret = pthread_mutex_unlock(&cancel_mutex);
 		if (ret)
 			vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
-		handle_ast_keys(h);
 		ret = pthread_mutex_lock(&io_mutex);
 		if (ret)
 			vms_fatal_exit("pthread_mutex_lock failed: %d", ret);
@@ -230,7 +254,6 @@ static void vms_input_thread(void *n, int h)
 		ret = pthread_mutex_unlock(&io_mutex);
 		if (ret)
 			vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
-		handle_ast_keys(h);
 		if (vms_ret == SS$_CANCEL ||
 		    vms_ret == SS$_ABORT)
 			continue;
@@ -258,17 +281,43 @@ static void vms_input_thread(void *n, int h)
 	}
 }
 
+static void tty_passthru_mode(int enable)
+{
+	struct _iosb iosb;
+	struct tt_mode mode;
+	int ret;
+
+	ret = sys$qiow(0, input_channel, IO$_SENSEMODE, &iosb, 0, 0, &mode, sizeof(mode), 0, 0, 0, 0);
+	if (!$VMS_STATUS_SUCCESS(ret)) {
+		error("IO$_SENSEMODE failed: %d", ret);
+		return;
+	}
+
+	if (enable) mode.extended |= TT2$M_PASTHRU;
+	else mode.extended &= ~TT2$M_PASTHRU;
+
+	ret = sys$qiow(0, input_channel, IO$_SETMODE, &iosb, 0, 0, &mode, sizeof(mode), 0, 0, 0, 0);
+	if (!$VMS_STATUS_SUCCESS(ret)) {
+		error("IO$_SETMODE failed: %d", ret);
+		return;
+	}
+}
+
 int setraw(int ctl, int save)
 {
 	int ret;
 	if (io_raw > 0)
 		return 0;
+
+	do_signal(SIGINT, SIG_IGN);
+
 	ret = sys$assign(&input_channel_desc, &input_channel, 0, NULL);
 	if (!$VMS_STATUS_SUCCESS(ret))
 		vms_fatal_exit("sys$assign failed: %d", ret);
 	io_raw = 1;
-	vms_key_handler(IO$M_CTRLCAST, ctrl_c_ast);
-	vms_key_handler(IO$M_CTRLYAST, ctrl_y_ast);
+
+	tty_passthru_mode(1);
+
 	ret = pthread_mutex_unlock(&io_mutex);
 	if (ret)
 		vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
@@ -296,9 +345,10 @@ void setcooked(int ctl)
 	ret = pthread_mutex_unlock(&cancel_mutex);
 	if (ret)
 		vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
+
+	tty_passthru_mode(0);
+
 	io_raw = 0;
-	vms_key_handler(IO$M_CTRLCAST, NULL);
-	vms_key_handler(IO$M_CTRLYAST, NULL);
 	ret = sys$dassgn(input_channel);
 	if (!$VMS_STATUS_SUCCESS(ret))
 		vms_fatal_exit("sys$dassgn failed: %d", ret);
@@ -306,7 +356,6 @@ void setcooked(int ctl)
 
 int get_input_handle(void)
 {
-	unsigned int disable_ctrl_mask = LIB$M_CLI_CTRLT | LIB$M_CLI_CTRLY;
 	int ret;
 	if (input_handle >= 0)
 		return input_handle;
@@ -319,7 +368,6 @@ int get_input_handle(void)
 	ret = pthread_mutex_lock(&io_mutex);
 	if (ret)
 		vms_fatal_exit("pthread_mutex_lock failed: %d", ret);
-	lib$disable_ctrl(&disable_ctrl_mask, &old_ctrl_mask);
 	vms_thread_high_priority = 1;
 	input_handle = start_thread(vms_input_thread, NULL, 0, 0);
 	if (input_handle == -1)
@@ -922,7 +970,7 @@ int vms_close(int fd)
 	return close(fd);
 }
 
-int c_pipe(int *fd)
+int vms_pipe(int fd[2])
 {
 #ifdef VMS_VIRTUAL_PIPE
 	return vpipe_create(fd);
@@ -955,10 +1003,8 @@ int vms_read(int fd, void *buf, size_t size)
 	int r = vpipe_read(fd, buf, size);
 	if (r != -2) return r;
 #endif
-#ifndef OPENVMS_64BIT
-	return read(fd, buf, size);
-#else
-	{
+#ifdef OPENVMS_64BIT
+	if ((my_uintptr_t)buf + size >= 0x80000000U) {
 		int r;
 		unsigned char static_buffer[BOUNCE_BUFFER_SIZE];
 		unsigned char *bounce_buffer = NULL;
@@ -975,6 +1021,7 @@ int vms_read(int fd, void *buf, size_t size)
 		return r;
 	}
 #endif
+	return read(fd, buf, size);
 }
 
 int vms_write(int fd, const void *buf, size_t size)
@@ -983,10 +1030,8 @@ int vms_write(int fd, const void *buf, size_t size)
 	int r = vpipe_write(fd, buf, size);
 	if (r != -2) return r;
 #endif
-#ifndef OPENVMS_64BIT
-	return write(fd, buf, size);
-#else
-	{
+#ifdef OPENVMS_64BIT
+	if ((my_uintptr_t)buf + size >= 0x80000000U) {
 		int r;
 		unsigned char static_buffer[BOUNCE_BUFFER_SIZE];
 		unsigned char *bounce_buffer = NULL;
@@ -1003,8 +1048,143 @@ int vms_write(int fd, const void *buf, size_t size)
 		return r;
 	}
 #endif
+	return write(fd, buf, size);
 }
 
+#ifdef VMS_ADVANCED_HEAP
+
+struct region {
+	unsigned char *start;
+	size_t len;
+	int allocated;
+};
+
+static struct region *regions = NULL;
+static int n_used_regions = 0;
+static int n_allocated_regions = 0;
+static unsigned char *max_expreg = NULL;
+
+static int expand_regions(void)
+{
+	struct region *regs;
+	if (n_used_regions < n_allocated_regions)
+		return 0;
+	if (n_allocated_regions >= MAXINT / 2 / sizeof(struct region))
+		return -1;
+	regs = realloc(regions, (n_allocated_regions + 1) * sizeof(struct region));
+	if (!regs)
+		return -1;
+	regions = regs;
+	n_allocated_regions++;
+	return 0;
+}
+
+void *virtual_alloc(size_t len)
+{
+	int i;
+	struct va_range expr, cre_range;
+	int vms_ret;
+
+search_again:
+	for (i = 0; i < n_used_regions; i++) {
+		if (regions[i].allocated)
+			continue;
+		if (regions[i].len < len)
+			continue;
+		if (regions[i].len == len) {
+			regions[i].allocated = 1;
+			goto ret_st;
+		}
+		if (expand_regions())
+			return NULL;
+		memmove(&regions[i + 1], &regions[i], (n_used_regions - i) * sizeof(struct region));
+		n_used_regions++;
+		regions[i].len = len;
+		regions[i].allocated = 1;
+		regions[i + 1].start = regions[i].start + len;
+		regions[i + 1].len -= len;
+
+ret_st:
+		cre_range.va_range$ps_start_va = regions[i].start;
+		cre_range.va_range$ps_end_va = regions[i].start + len - 1;
+		vms_ret = links_cretva(&cre_range, &expr, 0);
+		if (vms_ret != SS$_NORMAL) {
+			/*fprintf(stderr, "cretva failed\n");*/
+			regions[i].allocated = 0;
+			return NULL;
+		}
+
+		/*fprintf(stderr, "allocated: %p,%p ; %p,%p\n", cre_range.va_range$ps_start_va, cre_range.va_range$ps_end_va, expr.va_range$ps_start_va, expr.va_range$ps_end_va);*/
+
+		/*fprintf(stderr, "found existing (%d)\n", n_used_regions);*/
+		/*memset(regions[i].start, 0, regions[i].len);*/
+
+		return regions[i].start;
+	}
+	if (expand_regions())
+		return NULL;
+	vms_ret = links_expreg(len >> 9, &expr, 0, 0);
+	if (vms_ret != SS$_NORMAL) {
+		return NULL;
+	}
+	/*fprintf(stderr, "expreg (%p,%p %p) (%d)\n", expr.va_range$ps_start_va, expr.va_range$ps_end_va, (void *)len, n_used_regions);*/
+	max_expreg = expr.va_range$ps_end_va;
+	regions[n_used_regions].start = expr.va_range$ps_start_va;
+	regions[n_used_regions].len = (unsigned char *)expr.va_range$ps_end_va - (unsigned char *)expr.va_range$ps_start_va + 1;
+	regions[n_used_regions].allocated = 1;
+	n_used_regions++;
+	if (len != regions[n_used_regions - 1].len) {
+		/*fprintf(stderr, "size mismatch\n");*/
+		regions[n_used_regions - 1].allocated = 0;
+		return NULL;
+	}
+	return expr.va_range$ps_start_va;
+}
+
+void virtual_free(void *ptr, size_t len)
+{
+	int i;
+	struct va_range del_range, del_ret;
+	int vms_ret;
+
+	for (i = 0; i < n_used_regions; i++) {
+		if (!regions[i].allocated)
+			continue;
+		if (regions[i].start == ptr && regions[i].len == len)
+			goto found;
+	}
+	vms_fatal_exit("alloaction %p, %p not found", ptr, (void *)len);
+found:
+	/*fprintf(stderr, "dealloc (%d - %p,%p) (%d)\n", i, ptr, (void *)len, n_used_regions);*/
+	regions[i].allocated = 0;
+	if (i < n_used_regions - 1 && !regions[i + 1].allocated && regions[i].start + regions[i].len == regions[i + 1].start) {
+		regions[i].len += regions[i + 1].len;
+		memmove(&regions[i + 1], &regions[i + 2], (n_used_regions - i - 2) * sizeof(struct region));
+		n_used_regions--;
+		/*fprintf(stderr, "front merge\n");*/
+	}
+	if (i && !regions[i - 1].allocated && regions[i - 1].start + regions[i - 1].len == regions[i].start) {
+		regions[i - 1].len += regions[i].len;
+		memmove(&regions[i], &regions[i + 1], (n_used_regions - 1 - i) * sizeof(struct region));
+		n_used_regions--;
+		/*fprintf(stderr, "back merge\n");*/
+	}
+
+	del_range.va_range$ps_start_va = ptr;
+	del_range.va_range$ps_end_va = (unsigned char *)ptr + (len - 1);
+	if (del_range.va_range$ps_end_va == max_expreg) {
+		del_range.va_range$ps_end_va = (unsigned char *)del_range.va_range$ps_end_va - page_size;
+		memset((unsigned char *)del_range.va_range$ps_end_va + 1, 0, page_size);
+	}
+	if (del_range.va_range$ps_start_va <= del_range.va_range$ps_end_va) {
+		vms_ret = links_deltva(&del_range, &del_ret, 0);
+		if (vms_ret != SS$_NORMAL)
+			vms_fatal_exit("deltva failed: %d", vms_ret);
+	}
+}
+
+#endif
+
 void init_os(void)
 {
 #if !defined(__VAX) && defined(__FEATURE_MODE_INIT_STATE) && defined(__FEATURE_MODE_CURVAL)
@@ -1014,10 +1194,10 @@ void init_os(void)
 #if defined(OPENVMS_64BIT)
 	{
 		int i;
-		unsigned char **new_argv = malloc((g_argc + 1) * sizeof(unsigned char *));
+		char **new_argv = malloc((g_argc + 1) * sizeof(char *));
 		if (!new_argv) vms_fatal_exit("can't allocate argv");
 		for (i = 0; i <= g_argc; i++)
-			new_argv[i] = (unsigned char *)(my_intptr_t)(((unsigned *)g_argv)[i]);
+			new_argv[i] = (char *)(my_intptr_t)(((unsigned *)g_argv)[i]);
 		g_argv = new_argv;
 	}
 #endif
@@ -1062,6 +1242,7 @@ void init_os(void)
 void os_seed_random(unsigned char **pool, int *pool_size)
 {
 	struct history_item *hi;
+	struct list_head *lhi;
 	DIR *dir;
 	int n, h;
 
@@ -1080,14 +1261,14 @@ void os_seed_random(unsigned char **pool, int *pool_size)
 	 * any URLs in the history.
 	 */
 	n = 0;
-	foreach(hi, goto_url_history.items) {
+	foreach(struct history_item, hi, lhi, goto_url_history.items) {
 		SHA_CTX ctx;
 		unsigned char result[SHA_DIGEST_LENGTH];
 		unsigned char sum;
 		int i;
 
 		SHA1_Init(&ctx);
-		SHA1_Update(&ctx, hi->d, strlen(cast_const_char hi->d));
+		SHA1_Update(&ctx, hi->str, strlen(cast_const_char hi->str));
 		SHA1_Final(result, &ctx);
 
 		sum = 0;
@@ -1134,8 +1315,6 @@ void os_seed_random(unsigned char **pool, int *pool_size)
 
 void terminate_osdep(void)
 {
-	if (old_ctrl_mask)
-		lib$enable_ctrl(&old_ctrl_mask, NULL);
 }
 
 #else
